import logging

logger_manager = logging.getLogger('pipeline_engine.manager')
logger_manager_lock = logging.getLogger('pipeline_engine.manager.locks')
logger_node = logging.getLogger('pipeline_engine.node')
logger_decorators = logging.getLogger('pipeline_engine.decorators')
logger_parallelism = logging.getLogger('pipeline_engine.parallelism')
logger_visualization = logging.getLogger('pipeline_engine.visualization')
